local M = {}

local logTag = "raceUtils"

local function timeToString(t)
  local minutes = math.floor(t / 60);
  local seconds = t - (minutes * 60);
  return string.format("%02.0f:%05.2f", minutes, seconds)
end

local function loadPrefab(name, filePath) -- loads or reloads a prefab file
  if scenetree.findObject(name) then
    scenetree.findObject(name):delete()
  end
  local prefabObj = spawnPrefab(name, filePath, "0 0 0", "0 0 1", "1 1 1", true)
  if prefabObj then
    scenetree.MissionGroup:add(prefabObj)
    log("I", logTag, "Loaded prefab: "..name)
    return true
  else
    return false
  end
end

local function unloadPrefab(name) -- unloads a prefab, by name
  if scenetree.findObject(name) then
    scenetree.findObject(name):delete()
    log("I", logTag, "Unloaded prefab: "..name)
  end
end

local function setVehPaint(id, paint1, paint2, paint3) -- sets a vehicle's paint colors
  local obj = be:getObjectByID(id)
  if not obj then return end

  local layers = {paint1 or {}, paint2 or {}, paint3 or {}}
  local colorFields = {"color", "colorPalette0", "colorPalette1"}
  for i, p in ipairs(layers) do
    if p.baseColor then
      local metallicPaint = vehicleMetallicPaintString(p.metallic, p.roughness, p.clearcoat, p.clearcoatRoughness)
      obj[colorFields[i]] = ColorF(p.baseColor[1], p.baseColor[2], p.baseColor[3], p.baseColor[4]):asLinear4F()
      obj:setField("metallicPaintData", tostring(i - 1), metallicPaint)
    end
  end
end

local function setupMarkers(wpList, lapCount) -- sets up the race marker system
  local markers = require("scenario/race_marker")
  markers.init()
  local raceData = {waypoints = {}, totalLaps = lapCount or 1, currIdx = 0, currLap = 1}

  for _, wp in ipairs(wpList) do
    local obj = scenetree.findObject(wp)
    table.insert(raceData.waypoints, {name = wp, pos = vec3(obj:getPosition()), radius = vec3(obj:getScale()).z})
  end
  markers.setupMarkers(raceData.waypoints)
  log("I", logTag, "Created custom race markers data")

  return markers, raceData
end

local function updateMarkers(markers, raceData) -- updates the race data and marker modes (should be triggered when a waypoint event fires)
  local wpCount = #raceData.waypoints
  raceData.currIdx = raceData.currIdx + 1
  if raceData.currIdx > wpCount then
    raceData.currLap = raceData.currLap + 1

    if raceData.currLap > raceData.totalLaps then
      markers.setModes({})
      return markers, raceData
    else
      raceData.currIdx = 1
    end
  end

  local modes = {}
  local wp1, wp2 = raceData.waypoints[raceData.currIdx], raceData.waypoints[raceData.currIdx + 1]
  if wp1 then
    if raceData.currIdx < wpCount then
      modes[wp1.name] = "default"
    else
      if raceData.currLap < raceData.totalLaps then
        modes[wp1.name] = "lap"
      else
        modes[wp1.name] = "final"
      end
    end
  end
  if wp2 then
    modes[wp2.name] = "next"
  end

  markers.setModes(modes)
  return markers, raceData
end

local function finishScreen(title, text, failed, timer, statsData) -- custom finish screen UI
  local data = {}
  statsData = statsData or {}
  local mainStat = statsData.score or {}
  local medals = {"bronze", "silver", "gold"}
  local medalScore = mainStat.points or 0
  local buttonsTable = {}
  table.insert(buttonsTable, {label = "ui.common.retry", cmd = "extensions.hook('onResetGameplay')"})
  table.insert(buttonsTable, {label = "ui.dashboard.scenarios", cmd = "openScenarios"})
  table.insert(buttonsTable, {label = "ui.common.menu", cmd = "extensions.hook('onRaceToolRequest')"})

  data.buttons = buttonsTable
  data.title = title
  data.text = text
  data.time = timer and timeToString(timer)

  data.stats = statsData
  data.overall = {
    failed = failed,
    medal = medals[medalScore] or "wood",
    points = mainStat.points or nil,
    maxPoints = mainStat.maxPoints or nil
  }

  log("I", logTag, "Processing finish screen")
  bullettime.set(0.125)
  guihooks.trigger("ChangeState", {state = "scenario-end", params = {stats = data, portrait = {}}})
end

local function checkCondition(cond, stats) -- returns true if an achievement condition statement checks out
  -- cond = {value1, comparison, value2} -- values can also be stats keys
  local valid = false
  local var1 = type(cond[1]) == "number" and cond[1] or stats[cond[1]]
  local var2 = type(cond[3]) == "number" and cond[3] or stats[cond[3]]

  if var1 and var2 then
    if cond[2] == "lte" then
      if var1 <= var2 then
        valid = true
      end
    elseif cond[2] == "gte" then
      if var1 >= var2 then
        valid = true
      end
    else
      if var1 == var2 then
        valid = true
      end
    end
  end

  return valid
end

M.timeToString = timeToString
M.loadPrefab = loadPrefab
M.unloadPrefab = unloadPrefab
M.setVehPaint = setVehPaint
M.setupMarkers = setupMarkers
M.updateMarkers = updateMarkers
M.finishScreen = finishScreen
M.checkCondition = checkCondition

return M